// ************** MFSK_TX_01.C - MFSK Encoder - 27/02/2009 *************** 
//
// All code copright (c) 2009 Steve Randall, Laurence Blaxter 
//
//	
//
//          +-----------+    +-----------+    +---------------+    +-------------+    +------+    +-----------+
//  Data -->| Idle      |--->| Varicoder |--->| Convolutional |--->| Interleaver |--->| Grey |--->| Modulator |
//          | Insertion |    +-----------+ ^  | FEC coder     | ^  +-------------+    | Code | ^  +-----------+
//          +-----------+                  |  +---------------+ |                     +------+ |
//                                         |                    |                              |
//                                         |                    |                              |
//                                      8 Bits              16 Bits                         4 Bits
//
//
// General methodology - most significant bits are transmitted 1st
//
// idle is NULL + lots of zeros

/*
 * Compile with: gcc mfsk_tx_01.c -lsndfile -lm -o mfsk_tx
 *
 */

// stubs etc.

#include <avr.h>
#include 

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "lookups.h" // header etc.
#include "macros.h"


// *************************** MFSK tx code propper ***************************

#define u08 unsigned char
#define u16 unsigned short
#define u32 unsigned long
#define s08 char

// prototypes of functions that MFSK code calls
void EnableTX(int);   // turn the transmitter on (1) or off (0)
void Modulate(u08);   // Modualte the encoded data to air

static u08 cycbuf[16]; 
static u08 cycptr;   // points at last position in buffer 
static u08 encstate; // encoder state - rolling input
static u32 varibuf;  // large enough to allow for worst case varicode lengths 
static s08 varibits; // count of bits in buffer

// MFSK_Init() - initialize data structures

void MFSK_Init (void)
{
	int i;
	for (i = 0; i < 16; i++)
		cycbuf[i] = 0;
	cycptr = 0;
	encstate = 0;
	varibuf = 0;
	varibits = 0;
}

// passed a nibble (4 bits)
// calls DAC to transmit the bits

void GrayEncode(u08 arg)		// gray encodes a nibble
{
	Modulate(GreyTab[arg >> 4]);	// high nibble
	Modulate(GreyTab[arg & 0x0F]);   // low nibble
}


// Interleave() - Bit Interveiver 
// implments the equivelent of 10 sequential 4 bit diagonal interleavers
// equivelent is an offset of 10 rather than 1 (i.e. offset of 41 bits rather than 5 bits)
//
// Example Input bits: "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789ABCDE..."
//
// |A|E I M Q U Y 2 6 A E I M Q U Y 2 6 A E I M Q U Y 2 6 A E I M Q U Y 2 6 
//  B F J N R V Z 3 7 B|F|J N R V Z 3 7 B F J N R V Z 3 7 B F J N R V Z 3 7 
//  C G K O S W 0 4 8 C G K O S W 0 4 8 C G|K|O S W 0 4 8 C G K O S W 0 4 8 
//  D H L P T X 1 5 9 D H L P T X 1 5 9 D H L P T X 1 5 9 D H L|P|T X 1 5 9 
//
//  0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 
//  1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 
//
// Output bits: AFKPEJOT INSXMRW1 
//
// uses a 16 byte (128bit) cyclic buffer
// feed a byte in - get a interleved byte out.
// needs 16 bytes into this function to flush everything

void Interleave(u08 input) 
{ 
	u08 output; // no need to initiallize as fully written
        cycbuf[cycptr & 0x0F] = input; // cycptr now points at the buffer base.
        output = (cycbuf[cycptr & 0x0F] & 0x88) | (cycbuf[(cycptr + 5) & 0x0F] & 0x44) | 
                (cycbuf[(cycptr + 10) & 0x0F] & 0x22) | (cycbuf[(cycptr + 15) & 0x0F] & 0x11);
        --cycptr; // cycptr now points at where the next byte will go on next call
	GrayEncode(output); // bit significance preserved.
}


// FEC_Encode() - encode a byte into 2 bytes using r=1/2, k=7 convolution
// needs 9 bytes into this function to flush interleaver

void FEC_Encode(u08 input)
{
	u16 output; // no need to initialize as fully written
	int i;
	for(i = 7; i >= 0; i--)
	{ // for each bit in the input
	     encstate = (encstate << 1) | ((input >> i) & 1); // shift the bits in - msb 1st
	     output = (output << 2) | (u16) PolyTab[encstate & 0x7F]; // 7 bit table
	}
	Interleave((u08)(output >> 8));	// hi byte
	Interleave((u08)(output & 0x00FF)); 	// lo byte
}



void VariEncode(u08 ascii) //varicodes input character and calls FEC when it has enough Input
{
	u08 length;
	u16 value;
	value = VariTab[ascii]; //the upper 4 msb tell us how many valid bits there are for this character
	length = (u08)(value >> 12);	// extract length
	value &= 0x0FFF;		// strip length to just get bits 
	varibuf <<= length;	// make room at bottom
	varibuf |= (u32)value;	// add new varicode character in at the bottom
	varibits += length;	// count total bits in buffer
	while(varibits >= 8)
	{ // while there are complete bytes transmit them
		varibits -= 8;
		FEC_Encode((u08)(varibuf >> varibits));
	}
	// 0 to 7 bits remain
}

// VariFlush() - flush varicode, FEC and interleave buffers with zeros
// results in idle being sent (continious lowest tone)
//
// transmission starts with at least 8 symbols of the lowest code
// transmission ends with at least 4 symbols of the lowest code

void VariFlush(void) 
{
	int i;
	if (varibits > 0)
	{ // some varicode to flush
	     varibits -= 8;
	     varibuf <<= varibits; // shift in 0s
	     FEC_Encode((u08)(varibuf & 0x00FF)); // send remaining bits
	     varibits = 0;
	}
	for (i = 0; i < 9; i++)
	{
	     FEC_Encode(0); // send enough to flush all buffering
	}
}


// MFSK_loop() - detect start and end of transmission
// Insert pre and post message idle characters
// encode and transmit message
void MFSK_loop(int loop)
{
	u08 ch;
	int i;
	do 
	{
	    ch = MFSK_getch(); // collect next byte if ready - NULL if not
	    if (ch != 0)
	    { // 1st character of run
			EnableTX(1);		// power on
			MFSK_Init();	// re-initilize buffers
			VariFlush();	// and send idle preabble
			do
			{
				VariEncode(ch);
				ch = MFSK_getch(); // get next byte if ready
			}
			while(ch != 0);
			VariFlush();	// send idle postamble
			EnableTX(0);			// power off
	    }
	}
	while(loop); // loop or not
}
